/*
  DAQLIB: set of C library routines for programming a
  National Instruments AT-MIO-16E-1 DAQ card (it will
  likely work with other E-series ISA boards as well).

  It is relatively compact (doesn't use any devices) and
  this ideal for real-time applications.  There are some
  limitations, such as only supporting a single input
  and output channel, and many potentially configurable
  options are hard coded.  I/O is only supported in immediate
  mode, though you can use the clock to trigger an IRQ and
  set up periodic acquisiton yourself.  If you want a more fully 
  featured driver, I suggest you check out COMEDI, at 
  http://stm.lbl.gov/comedi/

  author: Rob Butera - butera@helix.nih.gov

  28 January 1999

  Copyright (C) 1999  Robert J. Butera, Jr.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  Or visit the following URL: http://www.fsf.org/copyleft/gpl.html

  USAGE NOTES: THIS IS ALL THE DOCS FOR NOW!

  It is HIGHLY reccomended to consult the following publications
  from National Instruments:

  AT-MIO E-series register level programming manual
  DAQ-STC technical reference manual

  Both are available from National Instruments, PDF files
  are available at their WWW site. For more information,
  visit the following URL: 

  http://www.natinst.com/support/trouble/daq/program/reg_prog.htm

  This IS the documentation.  All programs must be linked with
  both daqlib.c and eserfnct.c.

  FILES:

  daqlib.c    This file: application level routines
  eserfnct.c  Low-level board I/O routines.
  daqlib.h    header files - application should include this
  eserllp.h,daqscale.h other header files used by daqlib.c

  
  init_board() ;    initialize board - call this FIRST!

  Note: the #define NI_BASE_ADDRESS below must be set to your
  board address.  You will probably need to use the isapnptools
  program to set the address of your board. 

  init_ai(value) ;  initialize AI channel.  Only one channel is
                    supported.  VALUE is a short that sets the
		    dynamic input range of the channel, according
		    to the following table:

		    value   gain   input range (+/- this voltage)
		    000      0.5       10V
		    001        1        5V
		    010        2      2.5V
		    011        5        1V
		    100       10     500mV
		    101       20     250mV
		    110       50     100mV
		    111      100      50mV

		    We assume bipolar input.  With a little reading,
		    you can easily change this.

  init_ao() ;       initialize board for analog putput

  get_ai_raw() ;    returns a SHORT signed 12bit value (-2048 to 2047)
                    of the raw analog input

  get_ai_voltage() ; returns a scaled FLOAT of the analog input voltage in V

  put_ao_raw(value) ; write to analog output with a raw 12bit value (SHORT)

  put_ao_voltage(value) ; write to analog input with a voltage value (FLOAT)
                         an output range of +/- 10V is assumed.

  Clear_FIFO() ;    clears input FIFO.  


  config_dio( value ) ;  configure digital output.  Bits 0-7 specify
                         if bit is for output (1) or input (0).
  put_do( value );       write digital output (SHORT)
  get_di() ;             read digital input (returns SHORT)

  The AT-MII-16E-1 has two internal 20 MHz 24bit counters.
  We use counter 0 as a user-interactive timer, and counter 1
  to generate a periodic clock signal that triggers an ISA IRQ.

  init_timer() ;   initialize timer
  reset_timer() ;  reset timer to zero
  start_timer() ;  start timer counting up
  start_and_reset_timer() ;  reset then restart timer.
  stop_timer() ;   stop timer
  LONG read_timer() ; read timer (returns 24 bit long) ;
   
  For the CLOCK, the #define STC_INTERRUPT below specifies the DAQ-STC
  interrupt.  This maps to a corresponding ISA IRQ according to the
  following table:

  DAQ-STC interrupt   IRQ
    0                  3
    1                  4
    2                  5
    3                  7
    4                 10
    5                 11
    6                 12
    7                 15

  init_clock(count) ;  COUNT is a 16bit short specifying the number
                       of 20MHz (0.05 usec) intervals for the clock
		       period.  For example, use 500 to generate a
		       40 kHz signal (20MHz/40kHz=500).  Since the
		       counter is 24bit, this could be modified to
		       allow even longer signals (65535 equals about
		       300 Hz -the slowest the count will go), 
		       I just didn't write the call that way ...

  start_clock() ;      Start the clock
  stop_clock() ;       Stop the clock
  get_daq_irq() ;      Returns the IRQ the DAQ is using.  That way
                       the application program doesn't have to know.
		       Simple returns stc2irq[STC_INTERRUPT]
  ack_daq_interrupt() ; Clears the DAQ interrupt.


   This code borrowed heavily from numerous National Instruments
   example files.
*/

#include "eserrlp.h"
#include "daqscale.h"

/* global value - channel gain setting */

short channel_gain ;

/* global lookup table for input range settings */

const float vrange[] = { 20.0, 10.0, 5.0, 2.0, 1.0, 0.500, 0.200, 0.100 } ; 

/* global lookup table for mapping DAQ-STC interrupts to ISA IRQs */

const short stc2irq[] = {3,4,5,7,10,11,12,15} ;

/* global value - timer arming (needed for reset) */

char timer_is_armed ;

#define AI_CHANNEL 0
#define AO_CHANNEL 0

#define STC_INTERRUPT 2
#define NI_BASE_ADDRESS 0x100

/* ******************** INIT_BOARD *********************** */

/* NOT NEEDED */

short init_board(void) {

  return ( Set_Board_Base_Address(NI_BASE_ADDRESS) ) ;

}

/* ******************** CLEAR_FIFO *********************** */

void Clear_FIFO(void)
{
  /*
  * Writing to register Write_Strobe_1_Register with address 83.
  *    Write_Strobe_1 <= 1
  * New pattern = 0x0001
  */
  DAQ_STC_Windowed_Mode_Write(Write_Strobe_1_Register,0x0001);
  return;
} ;

/* 
   for now, channel not specified. 
   most of this mindlessly modified from aiex1.c

   gain value is as follows: for bipolar, corresponds to +/-

  *               000 0.5     10V
  *               001 1        5V
  *               010 2      2.5V
  *               011 5        1V
  *               100 10     500mV
  *               101 20     250mV
  *               110 50     100mV
  *               111 100     50mV
*/

/* ******************** INIT_AI *********************** */

void init_ai(short gain) {
  
  /* CONFIGURE BOARD */

  /*
  * Writing to register Write_Strobe_0_Register with address 82.
  *    Write_Strobe_0 <= 1
  * New pattern = 0x0001
  */
  DAQ_STC_Windowed_Mode_Write(Write_Strobe_0_Register,0x0001);
 
  /*
  * Writing to register Write_Strobe_1_Register with address 83.
  *    Write_Strobe_1 <= 1
  * New pattern = 0x0001
  */
  DAQ_STC_Windowed_Mode_Write(Write_Strobe_1_Register,0x0001);

  /*
  * Writing to register Config_Memory_High_Register with address 18.
  *    CM_Channel_Number <= channel_number (0)
  *    CM_Channel_Bank <= channel_bank (0)
  *    CM_Channel_Type <= channel_type_for_rev_b (3)
  * New pattern = 0x3000 NOPE! (NRSE)
  *
  * bits 14/13/12 - channel type (000 Calibration, 001 Diff, 010 NRSE, 011 RSE)
  * bits 5/4 Bank Select  - set to zero?
  * bits 3/2/1/0 Channel Select for current resource in scan list.
  * 
  */
  
  Board_Write(Configuration_Memory_High, (0x2000 | AI_CHANNEL) );

  /*
  * Writing to register Config_Memory_Low_Register with address 16.
  *    CM_Last_Channel <= last_channel (1)
  *    CM_Trigger <= trigger (0)
  *    CM_AI_Gain <= ai_gain (1)
  *    CM_AI_Polarity <= ai_polarity (0)
  *    CM_AI_Dither_Enable <= ai_dither_enable (0)
  * New pattern = 0x0101
  *
  * bitfields  15 LastChannel
  *            12 GenTrig (don't need for simple reading)
  *             9 Dither Enable
  *             8 unipolar(1)/bipolar(0)
  *           2-0 Input Gain Select 
  *               000 0.5
  *               001 1
  *               010 2
  *               011 5
  *               100 10
  *               101 20
  *               110 50
  *               111 100
  *
  * Us: LastChannel (only 1!), Gain=1
  */

  channel_gain = gain & 0x0007 ;

  Board_Write(Configuration_Memory_Low,( 0x8000 | channel_gain ) );

  /* PROGRAM THE DAQ-STC */

  /* *** MSC_Clock_Configure: timebase options */

  /*
  * Writing to register Clock_and_FOUT_Register with address 56.
  *    Slow_Internal_Timebase <= p->msc_slow_int_tb_enable (1)
  *    Slow_Internal_Time_Divide_By_2 <= p->msc_slow_int_tb_divide_by_2 (1)
  *    Clock_To_Board <= p->msc_clock_to_board_enable (1)
  *    Clock_To_Board_Divide_By_2 <= p->msc_clock_to_board_divide_by_2 (1)
  * New pattern = 0x1B00
  *
  * bit fields: 12 Slow_Internal_Time_Div_By_2   ZERO
  *             11 Slow_Internal_Timebase        ZERO
  *              9 Clock_To_Board/2              ZERO
  *              8 Clock_To_Board                ONE

  */
  DAQ_STC_Windowed_Mode_Write(Clock_and_FOUT_Register,0x0100);

  /* *** Clear_FIFO */

  Clear_FIFO() ;

  /* *** AI_Reset_All */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Reset <= 1
  * New pattern = 0x0001
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0001);

  /* IS THERE A REASON why there are two separate calls? */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0010);

  /*
  * Writing to register Interrupt_A_Ack_Register with address 2.
  *    AI_SC_TC_Error_Confirm <= 1
  *    AI_SC_TC_Interrupt_Ack <= 1
  *    AI_START1_Interrupt_Ack <= 1
  *    AI_START2_Interrupt_Ack <= 1
  *    AI_START_Interrupt_Ack <= 1
  *    AI_STOP_Interrupt_Ack <= 1
  *    AI_Error_Interrupt_Ack <= 1
  * New pattern = 0x3F80
  */
  DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,0x3F80);

  /*
  * Writing to register AI_Command_1_Register with address 8.
  *    AI_Command_1_Register <= 0
  * New pattern = 0x0000
  */
  DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,0x0000);

  /*
  * Writing to register AI_Mode_1_Register with address 12.
  *    Reserved_One <= 1
  *    AI_Start_Stop <= 1
  * New pattern = 0x000C
  */
  DAQ_STC_Windowed_Mode_Write(AI_Mode_1_Register,0x000C);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0100
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0100);

  /* *** AI_Board_Persionalize */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0010);

  /*
  * Writing to register Clock_and_FOUT_Register with address 56.
  *    AI_Source_Divide_By_2 <= p->ai_source_divide_by_2 (0)
  *    AI_Output_Divide_By_2 <= p->ai_output_divide_by_2 (1)
  * New pattern = 0x1B80
  *
  * bitfields 7 AI_Output_Divide_By_2
  *           6 AI_Source_Divide_By_2
  *           5 AO_Output_Divide_By_2
  *           4 AO_Source_Divide_By_2

  */
  DAQ_STC_Windowed_Mode_Write(Clock_and_FOUT_Register,0x0100);

  /*
  * Writing to register AI_Personal_Register with address 77.
  *    AI_CONVERT_Pulse_Timebase <= p->ai_convert_pulse_timebase (0)
  *    AI_CONVERT_Pulse_Width <= p->ai_convert_pulse_width (1)
  *    AI_FIFO_Flags_Polarity <= p->ai_fifo_flags_polarity (0)
  *    AI_LOCALMUX_CLK_Pulse_Width <= p->ai_localmux_clk_pulse_width (1)
  *    AI_AIFREQ_Polarity <= p->ai_aifreq_polarity (0)
  *    AI_SHIFTIN_Polarity <= p->ai_shiftin_polarity (0)
  *    AI_SHIFTIN_Pulse_Width <= p->ai_shiftin_pulse_width (1)
  *    AI_EOC_Polarity <= p->ai_eoc_polarity (0)
  *    AI_SOC_Polarity <= p->ai_soc_polarity (1)
  *    AI_Overrun_Mode <= p->ai_overrun_mode (1)
  * New pattern = 0xA4A0
  */
  DAQ_STC_Windowed_Mode_Write(AI_Personal_Register,0xA4A0);

  /*
  * Writing to register AI_Output_Control_Register with address 60.
  *    AI_CONVERT_Output_Select <= p->ai_convert_output_select (2)
  *    AI_SC_TC_Output_Select <= p->ai_sc_tc_output_select (3)
  *    AI_SCAN_IN_PROG_Output_Select <= p->ai_scan_in_prog_output_select (3)
  *    AI_LOCALMUX_CLK_Output_Select <= p->ai_localmux_clk_output_select (2)
  * New pattern = 0x032E
  */
  DAQ_STC_Windowed_Mode_Write(AI_Output_Control_Register,0x032E);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0100
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0100);


/* 
* Do this to access the first value in the configuration
* FIFO.
*/

  /* *** AI_Initialize_Configuration_Memory_Output */

  /*
  * Writing to register AI_Command_1_Register with address 8.
  *    AI_CONVERT_Pulse <= 1
  * New pattern = 0x0001
  */
  DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,0x0001); 


  /* *** AI_Board_Environmentalize */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0010);

  /*
  * Writing to register AI_Mode_2_Register with address 13.
  *    AI_External_MUX_Present <= 0
  * New pattern = 0x0000
  */
  DAQ_STC_Windowed_Mode_Write(AI_Mode_2_Register,0x0000);

  /*
  * Writing to register AI_Output_Control_Register with address 60.
  *    AI_EXTMUX_CLK_Output_Select <= p->ai_extmux_clk_output_select (0)
  * New pattern = 0x032E
  */
  DAQ_STC_Windowed_Mode_Write(AI_Output_Control_Register,0x032E);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0100
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0100);


/*
* Do this to enable or disable retriggering.
*/

  /* *** AI_Trigger_Signals */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0010);

  /*
  * Writing to register AI_Mode_1_Register with address 12.
  *    AI_Trigger_Once <= 1
  * New pattern = 0x000D
  */
  DAQ_STC_Windowed_Mode_Write(AI_Mode_1_Register,0x000D);

  /*
  * Writing to register AI_Trigger_Select_Register with address 63.
  *    AI_START1_Select <= 0
  *    AI_START1_Polarity <= 0
  *    AI_START1_Edge <= 1
  *    AI_START1_Sync <= 1
  * New pattern = 0x0060
  */
  DAQ_STC_Windowed_Mode_Write(AI_Trigger_Select_Register,0x0060);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0100
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0100);

  /* *** AI_Scan_Start */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0010);

  /*
  * Writing to register AI_START_STOP_Select_Register with address 62.
  *    AI_START_Select <= 0
  *    AI_START_Edge <= 1
  *    AI_START_Sync <= 1
  *    AI_START_Polarity <= 0
  * New pattern = 0x0060
  */
  DAQ_STC_Windowed_Mode_Write(AI_START_STOP_Select_Register,0x0060);


  /* *** AI_End_of_Scan */

 /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0100
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0100);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0010);

  /*
  * Writing to register AI_START_STOP_Select_Register with address 62.
  *    AI_STOP_Select <= p->ai_stop_select (19)
  *    AI_STOP_Edge <= 0
  *    AI_STOP_Polarity <= p->ai_stop_polarity (0)
  *    AI_STOP_Sync <= 1
  * New pattern = 0x29E0
  */
  DAQ_STC_Windowed_Mode_Write(AI_START_STOP_Select_Register,0x29E0);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0100
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0100);

  Clear_FIFO() ;
}


/* 
   AI ROUTINES 

   These are polling routines.  They could cause a hang if nothing
   is in the FIFO.  Maybe a future version should include a timeout
*/

/* ******************** GET_AI_RAW *********************** */

short get_ai_raw(void) {

  /* **** AI_Start_The_Acquisition 

  * Writing to register AI_Command_1_Register with address 8.
  *    AI_CONVERT_Pulse <= 1
  * New pattern = 0x0001
  */

  DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,0x0001);

  /* loop until something is in the buffer */

  while ( (DAQ_STC_Windowed_Mode_Read(AI_Status_1_Register) & 0x1000) == 0x1000 ) ;

  return ( Board_Read( ADC_FIFO_Data_Register ) ) ;
}

/* ******************** GET_AI_VOLTAGE *********************** */

/* NOTE: in a future version, we might just duplicate the above
   code, rather than call it, for speed reasons.  We'll see      */

float get_ai_voltage(void) {

  short ival ;
  float rval ;

  ival = get_ai_raw() ;

  rval = IVAL_2_VOLTAGE(ival) ;

  return rval ;
}


/* ****************** INIT_AO ********************** */

void init_ao(void) {

  /*
  * Configure the board.
  *
  * Writing to register AO_Configuration_Register with address 22.
  *    AO_DAC_Number_Select <= f2->f2_ao_channel_to_program (1)
  *    AO_DAC_Polarity <= p->dac_polarity (1)
  *    AO_DAC_Reference_Source <= p->dac_reference_source (0)
  *    AO_DAC_Low_Glitch <= p->dac_low_glitch (0)
  *    AO_DAC_Ground_Reference <= p->dac_ground_reference (26200)
  * New pattern = 0x0101
  *
  * bit assignments:
  *                   8  DACSelect (0/1)
  *                   3  GroundRef (1=connect ref for DACs to Ground)
  *                   2  ExtRef (1-external, 0-internal +10V)
  *                   1  ReGlitch (1=more uniform glitch, but higher energy)
  *                   0  Bipolar (1=-10V to 10V, 0=unipolar)
  */

  Board_Write(AO_Configuration_Register,(0x0001 | ( AO_CHANNEL << 8 )) );

  /* *** AO_Reset_All  **** */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AO_Configuration_Start <= 1
  * New pattern = 0x0020
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0020);

  /*
  * Writing to register AO_Command_1_Register with address 9.
  *    AO_Disarm <= 1
  * New pattern = 0x2000
  */
  DAQ_STC_Windowed_Mode_Write(AO_Command_1_Register,0x2000);

  /*
  * Writing to register Interrupt_B_Enable_Register with address 75.
  *    AO_BC_TC_Interrupt_Enable <= 0
  *    AO_START1_Interrupt_Enable <= 0
  *    AO_UPDATE_Interrupt_Enable <= 0
  *    AO_START_Interrupt_Enable <= 0
  *    AO_STOP_Interrupt_Enable <= 0
  *    AO_Error_Interrupt_Enable <= 0
  *    AO_UC_TC_Interrupt_Enable <= 0
  *    AO_FIFO_Interrupt_Enable <= 0
  * New pattern = 0x0000
  */
  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Enable_Register,0x0000);

  /*
  * Writing to register AO_Personal_Register with address 78.
  *    AO_BC_Source_Select <= 1
  *
  * NI Docs say that this bit MUST be set to 1 (zero is NOT supported)
  *
  * New pattern = 0x0010
  */
  DAQ_STC_Windowed_Mode_Write(AO_Personal_Register,0x0010);

  /*
  * Writing to register Interrupt_B_Ack_Register with address 3.
  *    AO_BC_TC_Trigger_Error_Confirm <= 1
  *    AO_BC_TC_Error_Confirm <= 1
  *    AO_UC_TC_Interrupt_Ack <= 1
  *    AO_BC_TC_Interrupt_Ack <= 1
  *    AO_START1_Interrupt_Ack <= 1
  *    AO_UPDATE_Interrupt_Ack <= 1
  *    AO_START_Interrupt_Ack <= 1
  *    AO_STOP_Interrupt_Ack <= 1
  *    AO_Error_Interrupt_Ack <= 1
  * New pattern = 0x3F98
  */
  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Ack_Register,0x3F98);

    /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AO_Configuration_Start <= 0
  *    AO_Configuration_End <= 1
  * New pattern = 0x0200
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0200);


  /* *** AO_Board_Personalize *** */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AO_Configuration_Start <= 1
  * New pattern = 0x0020
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0020);

  /*
  * Writing to register AO_Personal_Register with address 78.
  *    AO_Fast_CPU <= p->ao_fast_cpu (0)
  *    AO_UPDATE_Pulse_Timebase <= p->ao_update_pulse_timebase (0)
  *    AO_UPDATE_Pulse_Width <= p->ao_update_pulse_width (1)
  *    AO_DMA_PIO_Control <= p->ao_dma_pio_control (0)
  *    AO_AOFREQ_Polarity <= p->ao_dafreq_polarity (0)
  *    AO_TMRDACWR_Pulse_Width <= p->ao_tmrdacwr_pulse_width (1)
  *    AO_FIFO_Enable <= p->ao_fifo_enable (1)
  *    AO_FIFO_Flags_Polarity <= p->ao_fifo_flags_polarity (0)
  *    AO_Number_Of_DAC_Packages <= p->ao_number_of_dac_packages (0)
  * New pattern = 0x1430
  *
  * bit patterns 13 AO_Fast_CPU (1 for fast CPU, 0 for slow)
  * so we're setting it = patter: 0x3430
  *
  */
  DAQ_STC_Windowed_Mode_Write(AO_Personal_Register,0x3430);

  /*
  * Writing to register Clock_and_FOUT_Register with address 56.
  *    AO_Source_Divide_By_2 <= p->ao_source_divide_by_2 (0)
  *    AO_Output_Divide_By_2 <= p->ao_output_divide_by_2 (1)
  * New pattern = 0x1B20
  *
  * cut out the divide by two stuff - only selected bit is 8 (clocktoboard)
  */
  DAQ_STC_Windowed_Mode_Write(Clock_and_FOUT_Register,0x0100);

  /*
  * Writing to register AO_Output_Control_Register with address 86.
  *    AO_UPDATE_Output_Select <= p->ao_update_output_select (0)
  * New pattern = 0x0000
  */
  DAQ_STC_Windowed_Mode_Write(AO_Output_Control_Register,0x0000);

  /*
  * Writing to register AO_START_Select_Register with address 66.
  *    AO_AOFREQ_Enable <= p->ao_dafreq_enable (0)
  * New pattern = 0x0000
  */
  DAQ_STC_Windowed_Mode_Write(AO_START_Select_Register,0x0000);
  
  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AO_Configuration_Start <= 0
  *    AO_Configuration_End <= 1
  * New pattern = 0x0200
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0200);


  /* *** AO_LDAC_Source_And_Update_Mode *** */

  /*
   * Use this function to set the source and update mode for the LDAC<0..1> 
   * signals.
   */

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AO_Configuration_Start <= 1
  * New pattern = 0x0020
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0020);

  /*
  * Writing to register AO_Command_1_Register with address 9.
  *    AO_LDAC0_Source_Select <= p->ao_ldac0_source_select (0)
  *    AO_DAC0_Update_Mode <= p->ao_dac0_update_mode (0)
  *    AO_LDAC1_Source_Select <= p->ao_ldac1_source_select (0)
  *    AO_DAC1_Update_Mode <= p->ao_dac1_update_mode (0)
  * New pattern = 0x0014
  */
  DAQ_STC_Windowed_Mode_Write(AO_Command_1_Register,0x0000);

  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AO_Configuration_Start <= 0
  *    AO_Configuration_End <= 1
  * New pattern = 0x0200
  */
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,0x0200);
}


/* ************ PUT_AO_RAW ***************** */

void put_ao_raw( short ival ) {

  /*                         
  * Write to the DAC.
  *
  * Writing to register AO_DAC_1_Data_Register with address 26.
  *    AO_DAC_1_Data <= p->dac_data_point (1024)
  */

  Board_Write(AO_DAC_0_Data_Register, ( ival & 0x0FFF ) );
}                   

void put_ao_voltage( float voltage ) {

  short ival ;

  ival = VOLTAGE_2_IVAL(voltage) ;

  put_ao_raw( ival ) ;
}

/* ***************** DIGITAL I/O ********************** */

/* 0 for input, 1 for outout                              */
/* it is up to the user to remember what lines are in/out */

void config_dio( unsigned char config_byte )
{
  unsigned short dio_config_word ;

  dio_config_word = 0x0000 | config_byte ;

  DAQ_STC_Windowed_Mode_Write(DIO_Control_Register, dio_config_word) ;
}

void put_do(unsigned char dvalue)
{
  unsigned short dio_output_word ;

  dio_output_word = 0x0000 | dvalue ;
  DAQ_STC_Windowed_Mode_Write(DIO_Output_Register,dio_output_word);
}

unsigned char get_di(void)
{
  return ( 0x00FF & DAQ_STC_Windowed_Mode_Read(DIO_Parallel_Input_Register) ) ;
}


/* ******************* CLOCK FUNCTIONS ***************** */

/*
   CLOCK routines generate a periodic signal  
   They use counter 1 and are output to output.  In future,
   they will trigger Interrupt B

   TIMER routines are for timing and use counter 0
*/

void init_clock(short interval) {
  
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register, 0x0008) ;
  DAQ_STC_Windowed_Mode_Write(G1_Mode_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G1_Command_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G1_Input_Select_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G1_Autoincrement_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Enable_Register, 0x0000) ;

  DAQ_STC_Windowed_Mode_Write(G1_Command_Register, 0x0100) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Ack_Register, 0xC066) ;
  DAQ_STC_Windowed_Mode_Write(G1_Autoincrement_Register, 0x0000) ;

  /* now code from Simple_Gated_Event_Count */

  DAQ_STC_Windowed_Mode_Write(G1_Mode_Register, 0x0000) ;

  /* set load value and load into counter */

  DAQ_STC_Windowed_Mode_Write(G1_Load_A_Registers, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G1_Load_A_Registers+1, interval) ;
  DAQ_STC_Windowed_Mode_Write(G1_Command_Register, 0x0104) ;

  /* input select - use internal 20MHz clock, no gating */

  DAQ_STC_Windowed_Mode_Write(G1_Input_Select_Register, 0x0000) ;


  /* ... send TC to output (bit 8-9), load on TC (bit 12) */

  DAQ_STC_Windowed_Mode_Write(G1_Mode_Register, 0x1100) ;
 
  /* DEBUG ****************************************** */
  /*   pass output of TC to CTR1OUT pin on BNC-2090   */
  /* GPFO_1_Output_Enable, GPFO_1_Output_Select=TC(0) */

  DAQ_STC_Windowed_Mode_Write(Analog_Trigger_Etc_Register, 0x8000) ;
  
  /* ************************************************ */

  /* this works, now make count mode down and load on TC */
  
  /* count mode is DOWN */

  DAQ_STC_Windowed_Mode_Write(G1_Command_Register, 0x0104) ;

  /* setup to trigger interrupt B on IRQ 5 */

  /* 
     Interrupt_Output_Polarity = 0 (high)       bit 0 IntControlReg
     Interrupt_Output_On_3_Pins = 0 (disable)   bit 1 IntControlReg
     Interrupt_B_Output_Select = 2 (IRQ 5)      bits 12-14 IntControlReg
     Interrupt_B_Output_Enable (enabled)        bit 15 IntControlReg

     0xA000

     application note: within interupt routine, Interrupt must
     be both acknowledged (cleared) AND enabled for next interrupt
     to occur

     0x0200 (bit 9) enable TC interrupt on Interrupt_B_Enable_Register
  */

  DAQ_STC_Windowed_Mode_Write(Interrupt_Control_Register, (STC_INTERRUPT<<12)) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Enable_Register, 0x0200) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_Control_Register, ((STC_INTERRUPT<<12)|0x8000)) ;

  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Ack_Register,0xC066);
}

void start_clock(void) 
{ 
  /* unlike NI code, I don't see the point of setting the
     load bit here.  Besides, we want to arm/disarm at will */

  DAQ_STC_Windowed_Mode_Write(G1_Command_Register, 0x101) ; 
}

void stop_clock(void)
{
  DAQ_STC_Windowed_Mode_Write(G1_Command_Register, 0x0110) ; 
}

void ack_daq_interrupt(void)
{
  /* G1_TC Interrupt Ack */

  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Ack_Register, 0xC066) ;

  /*
  DAQ_STC_Windowed_Mode_Write(Interrupt_B_Enable_Register, 0x0200) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_Control_Register, (0x8000|(DAQ_IRQ<<12))) ;
  */

}  

short get_daq_irq(void) {
  return stc2irq[STC_INTERRUPT] ;
} 

/* ******************* TIMER funcs  **************************** */

/* based on NI sample routine GO_Reset_All */

void init_timer(void) {
  
  DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register, 0x0004) ;
  DAQ_STC_Windowed_Mode_Write(G0_Mode_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G0_Input_Select_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G0_Autoincrement_Register, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_A_Enable_Register, 0x0000) ;

  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x0100) ;
  DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register, 0xC060) ;
  DAQ_STC_Windowed_Mode_Write(G0_Autoincrement_Register, 0x0000) ;

  timer_is_armed = 0 ;

  /* now code from Simple_Gated_Event_Count */

  DAQ_STC_Windowed_Mode_Write(G0_Mode_Register, 0x0000) ;

  /* set load value (0) and load into counter */

  DAQ_STC_Windowed_Mode_Write(G0_Load_A_Registers, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G0_Load_A_Registers+1, 0x0000) ;
  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x0104) ;

  /* input select - use internal 20MHz clock, no gating */

  DAQ_STC_Windowed_Mode_Write(G0_Input_Select_Register, 0x0000) ;

  /* we're not using gating .... */

  DAQ_STC_Windowed_Mode_Write(G0_Mode_Register, 0x0000) ;

  /* count mode is up */

  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x0124) ;

  /* don't need interrupt A */

  DAQ_STC_Windowed_Mode_Write(Interrupt_A_Enable_Register, 0x0000) ;
}

void start_timer(void) 
{ 
  /* unlike NI code, I don't see the point of setting the
     load bit here.  Besides, we want to arm/disarm at will */

  timer_is_armed = 1 ;
  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x121) ; 
}

void start_and_reset_timer(void) 
{ 
  /* this code hopefully starts AND resets the timer all
     at once.  Let's see  */

  timer_is_armed = 1 ;
  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x134) ;
  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x121) ;
}


void stop_timer(void)
{
  timer_is_armed = 0 ;
  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x130) ; 
}

/* this "reset" just loads Reg A into the counter. Reg A was
   presumably set during init_timer()  Since writing to the
   command register, we need to know if the system is armed
   or not so that the arm/disarm bit is correctly set        */

/* hmmm.  resettng while armed doesn't work.  For now we'll
   call this a "feature" */

void reset_timer(void)
{
  /*
  short load_command ;
  
  if (timer_is_armed != 0) load_command = 0x125 ;
  else load_command = 0x134 ;

  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, load_command) ;
  */

 if (timer_is_armed==0)
   DAQ_STC_Windowed_Mode_Write(G0_Command_Register, 0x134) ;
}

unsigned long read_timer(void) 
{
  short load_command ;
  long rval ;

  /* why does the NI code write to the command register twice,
     once with the save_trace not set and once with it set?
     The first write doesn't seem necessary, but we'll do it
     anyway.  Also, why should the load  bit be set? */

  if (timer_is_armed != 0) load_command = 0x121 ;
  else load_command = 0x130 ;

  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, load_command) ; 

  load_command |= 0x0002 ;

  DAQ_STC_Windowed_Mode_Write(G0_Command_Register, load_command) ; 


  /* now the counter is saved --- read the save register */

  rval = DAQ_STC_Windowed_Mode_Read(G0_Save_Registers+1) +
         (((long) DAQ_STC_Windowed_Mode_Read(G0_Save_Registers)) << 16) ;

  return rval ;
}


